"""Antenna utils.

This module contains functions to generate antenna arrays, compute its baselines, 
perform aperture synthesis, obtain uv-coverage and get observations from sky models.

:Authors: Ezequiel Centofanti <ezequiel.centofanti@cea.fr>

"""

import numpy as np
import numpy.random as rnd

########################################
#      Generate antenna positions      #
########################################


def random_antenna_pos(E_lim=1000.0, N_lim=1000.0, U_lim=0.0):
    """Random antenna pos.

    Function to generate a random antenna location in ENU coordinates.
    Antenna lies in the range: [-E_lims/2, E_lims/2]x[-N_lims/2, N_lims/2]x[0, U_lims].

    Parameters
    ----------
    E_lim : float
        The east coordinate span width of the antenna position in meters.
    N_lim : float
        The north coordinate span width of the antenna position in meters.
    U_lim : float
        The up coordinate span width of the antenna position in meters.

    Returns
    -------
    antenna_pos : np.ndarray
        The antenna position in ENU coordinates.
    """
    # Return (x,y) random location for single dish
    return (
        rnd.random_sample(3) * np.array([E_lim, N_lim, U_lim])
        - np.array([E_lim, N_lim, 0.0]) / 2
    )


def circular_antenna_arr(n_antenna=3, r=300.0):
    """Circular antenna arr.

    Function to generate a circular antenna array. Antennas lie in a circumference
    of radius 'r' from the center [0,0] and are equally spaced.

    Parameters
    ----------
    n_antenna : int
        The number of antennas in the array.
    r : float
        The radius of the antenna array in meters.

    Returns
    -------
    antenna_arr : np.ndarray
        The antenna array positions in ENU coordinates.
    """
    # Return list of 'n' antenna locations (x_i, y_i) equally spaced over a 'r' radius circumference.
    return np.array(
        [
            [np.cos(angle) * r, np.sin(angle) * r, 0.0]
            for angle in [2 * np.pi / n_antenna * i for i in range(n_antenna)]
        ]
    )


def y_antenna_arr(n_antenna=5, r=500.0, alpha=0.0):
    """Y antenna arr.

    Function to generate a Y-shaped antenna array. Antennas lie equispaced in three radial arms
    of 120 degrees each.

    Parameters
    ----------
    n_antenna : int
        The number of antennas per arm.
    r : float
        The radius of the antenna array in meters.
    alpha : float
        The angle of the first arm.

    Returns
    -------
    antenna_arr : np.ndarray
        The antenna array positions in ENU coordinates.
    """
    # Return list of 'n' antenna locations (x_i, y_i) equispaced on three (120 deg) radial arms.
    step = r / n_antenna
    return np.array(
        [
            [
                np.array(
                    [
                        (i + 1) * step * np.cos(angle / 180 * np.pi),
                        (i + 1) * step * np.sin(angle / 180 * np.pi),
                        0.0,
                    ]
                )
                for i in range(n_antenna)
            ]
            for angle in [alpha, alpha + 120, alpha + 240]
        ]
    ).reshape((3 * n_antenna, 3))


def random_antenna_arr(n_antenna=3, E_lim=1000.0, N_lim=1000.0, U_lim=0.0):
    """Random antenna arr.

    Function to generate a random antenna array. Antennas lie randomly distributed
    in the range: [-E_lims/2, E_lims/2]x[-N_lims/2, N_lims/2]x[0, U_lims].

    Parameters
    ----------
    n_antenna : int
        The number of antennas in the array.
    E_lim : float
        The east coordinate span width of the antenna positions in meters.
    N_lim : float
        The north coordinate span width of the antenna positions in meters.
    U_lim : float
        The up coordinate span width of the antenna positions in meters.

    Returns
    -------
    antenna_arr : np.ndarray
        The antenna array positions in ENU coordinates.
    """
    # Return list of 'n' antenna locations (x_i, y_i) randomly distributed.
    return np.array([random_antenna_pos(E_lim, N_lim, U_lim) for i in range(n_antenna)])


def uni_antenna_array(
    n_antenna_E=32, n_antenna_N=32, E_lim=800.0, N_lim=800.0, U_lim=0.0
):
    """Uniform (grid) antenna arr.

    Function to generate a uniform antenna array. Antennas lie uniformely distributed
    in the range, and center on [0,0,U_lim]: [-E_lim/2, E_lim/2]x[-N_lim/2, N_lim/2]x[U_lim].
    Only allow to generate a grid at a fixed height U_lim (fixed Up component).

    Parameters
    ----------
    n_antenna_E : int
        The number of antennas in the North direction (vertical).
    n_antenna_N : int
        The number of antennas in the East direction (horizontal).
    E_lim : float
        The east coordinate span width of the antenna positions in meters.
    N_lim : float
        The north coordinate span width of the antenna positions in meters.
    U_lim : float
        The up coordinate span width of the antenna positions in meters.

    Returns
    -------
    antenna_arr : np.ndarray
        The antenna array positions in ENU coordinates.
    """
    # Return list of 'n_antenna_E x n_antenna_N' antenna locations (x_i, y_i) uniformly distributed, with Up component = U_lim.
    e = np.linspace(-E_lim / 2, E_lim / 2, n_antenna_E)
    n = np.linspace(-N_lim / 2, N_lim / 2, n_antenna_N)
    E, N = np.meshgrid(e, n)
    E_flat = E.flatten()
    N_flat = N.flatten()
    U_flat = np.zeros_like(E_flat) + U_lim
    array_grid = np.column_stack((E_flat, N_flat, U_flat))
    return array_grid


########################################
#  Compute baselines and uv sampling   #
########################################


def get_baselines(array):
    """Get baselines.

    Function to compute the baselines of an antenna array.

    Parameters
    ----------
    array : np.ndarray
        The antenna array positions.

    Returns
    -------
    baselines : np.ndarray
        The baselines of the antenna array.
    """
    # Get the baseline for every combination of antennas i-j.
    # Remove the i=j baselines: np.delete(array, list, axis=0) -> delete the rows listed on 'list' from array 'array'.
    return np.delete(
        np.array([antenna_i - antenna_j for antenna_i in array for antenna_j in array]),
        [(len(array) + 1) * n for n in range(len(array))],
        0,
    )


def ENU_to_XYZ(b_ENU, lat=35.0 / 180 * np.pi):
    """ENU to XYZ.

    Function to convert the baselines from East-North-Up (ENU) to XYZ coordinates.

    Parameters
    ----------
    b_ENU : np.ndarray
        The baselines in ENU coordinates.
    lat : float
        The latitude of the antenna array in radians.

    Returns
    -------
    X : np.ndarray
        The X coordinate of the baselines in XYZ coordinates.
    Y : np.ndarray
        The Y coordinate of the baselines in XYZ coordinates.
    Z : np.ndarray
        The Z coordinate of the baselines in XYZ coordinates.
    """
    # Compute baseline length, Azimuth and Elevation angles
    D = np.sqrt(np.sum(b_ENU**2, axis=1))
    A = np.arctan2(b_ENU[:, 0], b_ENU[:, 1])
    E = np.arcsin(b_ENU[:, 2] / D)
    # Compute the baseline in XYZ coordinates
    X = D * (np.cos(lat) * np.sin(E) - np.sin(lat) * np.cos(E) * np.cos(A))
    Y = D * np.cos(E) * np.sin(A)
    Z = D * (np.sin(lat) * np.sin(E) + np.cos(lat) * np.cos(E) * np.cos(A))

    return X, Y, Z


def XYZ_to_uvw(X, Y, Z, dec=30.0 / 180 * np.pi, ha=0.0, f=1420e6):
    """XYZ to uvw.

    Get the uvw sampling points from the XYZ coordinates given a
    source declination, hour angle and frequency.

    Parameters
    ----------
    X : np.ndarray
        The X coordinate of the baselines in XYZ coordinates.
    Y : np.ndarray
        The Y coordinate of the baselines in XYZ coordinates.
    Z : np.ndarray
        The Z coordinate of the baselines in XYZ coordinates.
    dec : float
        The declination of the source in radians.
    ha : float
        The hour angle of the source in radians.
    f : float
        The frequency of the observation in Hz.

    Returns
    -------
    u : np.ndarray
        The u coordinate of the baselines in uvw coordinates.
    v : np.ndarray
        The v coordinate of the baselines in uvw coordinates.
    w : np.ndarray
        The w coordinate of the baselines in uvw coordinates.
    """
    c = 299792458
    lam_inv = f / c
    u = lam_inv * (np.sin(ha) * X + np.cos(ha) * Y)
    v = lam_inv * (
        -np.sin(dec) * np.cos(ha) * X + np.sin(dec) * np.sin(ha) * Y + np.cos(dec) * Z
    )
    w = lam_inv * (
        np.cos(dec) * np.cos(ha) * X - np.cos(dec) * np.sin(ha) * Y + np.sin(dec) * Z
    )
    return u, v, w


def uv_track_multiband(
    b_ENU,
    lat=35.0 / 180 * np.pi,
    dec=35.0 / 180 * np.pi,
    track_time=0.0,
    t_0=0.0,
    n_times=1,
    f=1420e6,
    df=0.0,
    n_freqs=1,
):
    """Uv track multiband.

    Function to compute the uv sampling baselines for a given observation time and frequency range.

    Parameters
    ----------
    b_ENU : np.ndarray
        The baselines in ENU coordinates.
    lat : float
        The latitude of the antenna array in radians.
    dec : float
        The declination of the source in radians.
    track_time : float
        The duration of the tracking in hours.
    t_0 : float
        The initial tracking time in hours.
    n_times : int
        The number of time steps.
    f : float
        The central frequency of the observation in Hz.
    df : float
        The frequency range of the observation in Hz.
    n_freqs : int
        The number of frequency samples.

    Returns
    -------
    track : np.ndarray
        The uv sampling baselines listed for each time step and frequency.
    """
    # Compute the baselines in XYZ coordinates
    X, Y, Z = ENU_to_XYZ(b_ENU, lat)
    # Compute the time steps
    h = np.linspace(t_0, t_0 + track_time, n_times) * np.pi / 12
    # Compute the frequency range
    f_range = np.linspace(f - df / 2, f + df / 2, n_freqs)

    track = []
    for t in h:
        multi_band = []
        for f_ in f_range:
            u, v, w = XYZ_to_uvw(X, Y, Z, dec, t, f_)
            multi_band.append(np.array([u, v, w]))
        track.append(multi_band)
    track = np.array(track).swapaxes(-1, -2).reshape(-1, 3)

    return track


def combine_antenna_arr(arr1, arr2):
    """Combine antenna arr.

    Function to combine two antenna arrays.

    Parameters
    ----------
    arr1 : np.ndarray
        The first antenna array positions.
    arr2 : np.ndarray
        The second antenna array positions.

    Returns
    -------
    arr : np.ndarray
        The combined antenna array positions.
    """
    return np.concatenate((arr1, arr2), axis=0)


def load_antenna_enu_txt(path):
    """Load antenna txt.

    Function to load antenna ENU positions from a txt file.

    Parameters
    ----------
    path : str
        The path to the txt file.

    Returns
    -------
    antenna_arr : np.ndarray
        The antenna array positions.
    """
    return np.genfromtxt(path)



import numpy as np
import numpy.testing as npt

from argosim import antenna_utils as au


class TestAntennaUtils:

    antenna_lims = {"east": (-500.0, 500.0), "north": (-500.0, 500.0), "up": 0.0}

    circ_antenna_exp = np.array(
        [[300.0, 0.0, 0.0], [-150.0, 259.80762114, 0.0], [-150.0, -259.80762114, 0.0]]
    )
    circ_antenna_tolerance = 1e-7

    def test_random_antenna_pos_default(self):

        antenna_pos = au.random_antenna_pos()

        assert np.logical_and(
            antenna_pos[0] >= self.antenna_lims["east"][0],
            antenna_pos[0] <= self.antenna_lims["east"][1],
        )
        assert np.logical_and(
            antenna_pos[1] >= self.antenna_lims["north"][0],
            antenna_pos[1] <= self.antenna_lims["north"][1],
        )
        assert antenna_pos[2] == self.antenna_lims["up"]

    def test_circular_antenna_arr(self):

        circ_antenna_out = au.circular_antenna_arr()

        npt.assert_allclose(
            circ_antenna_out,
            self.circ_antenna_exp,
            atol=self.circ_antenna_tolerance,
            err_msg="Circular antenna outputs do not match.",
        )



"""Plot utils.

This module contains functions to plot the antenna array, beam, baselines, 
uv-coverage and sky models.

:Authors: Ezequiel Centofanti <ezequiel.centofanti@cea.fr>

"""

import matplotlib
import matplotlib.pyplot as plt
import numpy as np


def plot_beam(beam, pRng=(-0.1, 0.5), ax=None, fig=None):
    """Plot beam.

    Function to plot the synthesised beam image.

    Parameters
    ----------
    beam : np.ndarray
        The beam image.
    pRng : tuple
        The range of the colorbar.
    ax : matplotlib.axes.Axes
        The axis to plot the beam. For plotting on a specific subplot axis.
    fig : matplotlib.figure.Figure
        The figure to plot the beam. For plotting on a specific subplot axis.

    Returns
    -------
    None
    """
    # Imshow min and max values.
    zMin = np.nanmin(beam)
    zMax = np.nanmax(beam)
    zRng = zMin - zMax
    zMin -= zRng * pRng[0]
    zMax += zRng * pRng[1]
    if ax == None or fig == None:
        fig, ax = plt.subplots(1, 1)
    im = ax.imshow(np.fft.ifftshift(beam), vmin=zMin, vmax=zMax)
    fig.colorbar(im, ax=ax)
    if ax == None or fig == None:
        plt.show()


def plot_antenna_arr(array, ax=None, fig=None, title="Array"):
    """Plot antenna array.

    Function to plot the antenna array in ground coordinates.

    Parameters
    ----------
    array : np.ndarray
        The antenna array positions in the ground.
    ax : matplotlib.axes.Axes
        The axis to plot the antenna array. For plotting on a specific subplot axis.
    fig : matplotlib.figure.Figure
        The figure to plot the antenna array. For plotting on a specific subplot axis.
    title : str
        The title of the plot.

    Returns
    -------
    None
    """
    # Center the array
    array = array - np.mean(array, axis=0)
    if ax == None or fig == None:
        fig, ax = plt.subplots(1, 1)
    ax.scatter(array[:, 0], array[:, 1], s=20, c="gray")
    for i, txt in enumerate(range(1, len(array) + 1, 1)):
        ax.annotate(txt, (array[i, 0], array[i, 1]))
        ax.set_xlabel("x [m]")
        ax.set_ylabel("y [m]")
        x_lim = np.max(np.abs(array)) * 1.1
        y_lim = x_lim
        ax.set_xlim(-x_lim, x_lim)
        ax.set_ylim(-y_lim, y_lim)
        ax.set_aspect("equal", adjustable="box")
    ax.set_title(title)
    if ax == None or fig == None:
        plt.show()


def plot_baselines(baselines, ax=None, fig=None, ENU=False):
    """Plot baselines.

    Function to plot the baselines in uv-space.

    Parameters
    ----------
    baselines : np.ndarray
        The uv-space sampling positions.
    ax : matplotlib.axes.Axes
        The axis to plot the baselines. For plotting on a specific subplot axis.
    fig : matplotlib.figure.Figure
        The figure to plot the baselines. For plotting on a specific subplot axis.
    ENU : bool
        If True, plot the baselines in East-North-Up coordinates. Otherwise, plot in uv-space.

    Returns
    -------
    None
    """
    if ax == None or fig == None:
        fig, ax = plt.subplots(1, 1)
    if ENU:
        ax.set_xlabel("East [m]")
        ax.set_ylabel("North [m]")
        ax.set_title("Baselines")
        ax.scatter(baselines[:, 0], baselines[:, 1], s=0.4, c="gray")
        ax.set_xlim([np.min(baselines), np.max(baselines)])
        ax.set_ylim([np.min(baselines), np.max(baselines)])
    else:
        ax.set_xlabel(r"u(k$\lambda$)")
        ax.set_ylabel(r"v(k$\lambda$)")
        ax.set_title(r"uv-plane")
        ax.scatter(baselines[:, 0] / 1000, baselines[:, 1] / 1000, s=0.4, c="gray")
        ax.set_xlim([np.min(baselines) / 1000, np.max(baselines) / 1000])
        ax.set_ylim([np.min(baselines) / 1000, np.max(baselines) / 1000])
    ax.set_aspect("equal", adjustable="box")
    if ax == None or fig == None:
        plt.show()


def plot_sky(image, fov_size=(1.0, 1.0), ax=None, fig=None, title="Sky"):
    """Plot sky.

    Function to plot the sky model.

    Parameters
    ----------
    image : np.ndarray
        The sky model image.
    fov_size : tuple
        The field of view size in degrees.

    Returns
    -------
    None
    """
    if ax == None or fig == None:
        fig, ax = plt.subplots(1, 1)
    im = ax.imshow(
        image,
        extent=[-fov_size[0] / 2, fov_size[0] / 2, -fov_size[1] / 2, fov_size[1] / 2],
    )
    fig.colorbar(im, ax=ax)
    ax.set_xlabel("x [deg]")
    ax.set_ylabel("y [deg]")
    ax.set_title("{} ({}x{})".format(title, image.shape[0], image.shape[1]))
    if ax == None or fig == None:
        plt.show()


def plot_sky_uv(sky_uv, fov_size):
    """Plot sky uv.

    Function to plot the sky model in uv-space in logarithmic amplitud sale.

    Parameters
    ----------
    sky_uv : np.ndarray
        The sky model in uv-space.The inverse fourier transform of the sky model.

    Returns
    -------
    None
    """
    max_u = (180 / np.pi) * sky_uv.shape[0] / (2 * fov_size[0]) / 1000
    max_v = (180 / np.pi) * sky_uv.shape[1] / (2 * fov_size[1]) / 1000

    plt.figure(figsize=(10, 4))
    plt.subplot(121)
    plt.imshow(np.abs(sky_uv / 1000), extent=[-max_u, max_u, -max_v, max_v])
    plt.xlabel(r"$u(k\lambda$)")
    plt.ylabel(r"$v(k\lambda$)")
    plt.title("Amplitude")
    plt.subplot(122)
    plt.imshow(np.angle(sky_uv / 1000), extent=[-max_u, max_u, -max_v, max_v])
    plt.xlabel(r"$u(k\lambda$)")
    plt.ylabel(r"$v(k\lambda$)")
    plt.title("Phase")
    plt.show()


def plot_sampled_sky(sky_uv):
    """Plot sampled sky.

    Function to plot the sky model samples in uv-space according to the uv-coverage.

    Parameters
    ----------
    sky_uv : np.ndarray
        The sampled sky model in uv-space.

    Returns
    -------
    None
    """
    plt.imshow(np.abs(sky_uv) + 1e-3, norm=matplotlib.colors.LogNorm())
    plt.colorbar()
    plt.show()


def plot_uv_hist(baselines, bins=20, output_folder=None):
    """Plot uv histogram.

    Function to plot the histogram of the uv-sampling distribution.

    Parameters
    ----------
    baselines : np.ndarray
        The uv-space sampling positions.
    bins : int
        The number of bins for the histogram.
    output_folder : str
        The output folder to save the plot.

    Returns
    -------
    np.ndarray
        The histogram of the uv-sampling distribution.
    """
    # scale to kilo-lambda
    baselines = baselines / 1000

    cmap = matplotlib.colormaps["bone"]

    fig, ax = plt.subplots(1, 2, figsize=(11, 4))

    D = np.sqrt(np.sum(baselines[:, :2] ** 2, axis=1))
    baseline_hist = ax[0].hist(D, range=(0, np.max(D) * 1.1), bins=bins)

    n = baseline_hist[0]
    patches = baseline_hist[2]
    col = (n - n.min()) / (n.max() - n.min())
    for c, p in zip(col, patches):
        plt.setp(p, "facecolor", cmap(c))
    ax[0].set_title("Baselines histogram")
    ax[0].set_xlabel(r"UV distance $(k\lambda)$")
    ax[0].set_ylabel("Counts")
    ax[0].set_facecolor("palegoldenrod")
    ax[0].set_box_aspect(1)

    counts = np.flip(baseline_hist[0])
    r_list = baseline_hist[1]

    colors = cmap((counts / max(counts)))

    draw_back = plt.Circle((0.0, 0.0), 10 * r_list[-1], color="black", fill=True)
    ax[1].add_artist(draw_back)
    for color, r in zip(colors, np.flip(r_list[1:])):
        draw1 = plt.Circle((0.0, 0.0), r, color=color, fill=True)
        ax[1].add_artist(draw1)

    ax[1].scatter(baselines[:, 0], baselines[:, 1], color="yellow", s=1, alpha=0.3)

    fig.colorbar(
        plt.cm.ScalarMappable(
            cmap=cmap, norm=matplotlib.colors.Normalize(vmin=0, vmax=max(counts))
        ),
        ax=ax[1],
        orientation="vertical",
        label="Counts",
    )
    ax[1].set_aspect("equal")
    ax[1].set_xlim(-r_list[-1] * 1.1, r_list[-1] * 1.1)
    ax[1].set_ylim(-r_list[-1] * 1.1, r_list[-1] * 1.1)
    ax[1].set_title("Radial distribution")
    ax[1].set_xlabel(r"u $(k\lambda)$")
    ax[1].set_ylabel(r"v $(k\lambda)$")

    plt.suptitle(
        "UV sampling distribution",
        horizontalalignment="center",
        verticalalignment="top",
    )

    if output_folder is not None:
        plt.savefig(output_folder + "uv_hist.pdf")
    else:
        plt.show()

    return baseline_hist



"""Clean.

This module contains the functions to perform 
the Hogbom's clean algorithm on dirty observations.

:Authors: Ezequiel Centofanti <ezequiel.centofanti@cea.fr>

"""

import numpy as np

from argosim.data_utils import gauss_source


def shift_beam(beam, shift_x, shift_y):
    """Shift beam.

    Function to shift the beam image by a given amount of pixels in the
    x and y directions.

    Parameters
    ----------
    beam : np.ndarray
        The beam image.
    shift_x : int
        The shift in the x direction.
    shift_y : int
        The shift in the y direction.

    Returns
    -------
    beam_shift : np.ndarray
        The shifted beam image.
    """
    beam_shift = np.roll(beam, shift_x, axis=1)
    if shift_x < 0:
        beam_shift[:, shift_x:] = 0
    else:
        beam_shift[:, :shift_x] = 0

    beam_shift = np.roll(beam_shift, shift_y, axis=0)
    if shift_y < 0:
        beam_shift[shift_y:, :] = 0
    else:
        beam_shift[:shift_y, :] = 0
    return beam_shift


def find_peak(I):
    """Find peak.

    Function to find the peak of an image.

    Parameters
    ----------
    I : np.ndarray
        The image.

    Returns
    -------
    max_val : float
        The maximum value of the image.
    x_max : int
        The x coordinate of the maximum value.
    y_max : int
        The y coordinate of the maximum value.
    shift_x : int
        The shift in the x direction from the center of the image.
    shift_y : int
        The shift in the y direction from the center of the image.
    """
    y_max, x_max = int(np.argmax(I) / I.shape[0]), np.mod(np.argmax(I), I.shape[0])
    x_off, y_off = int(I.shape[0] / 2), int(I.shape[1] / 2)
    shift_x, shift_y = x_max - x_off, y_max - y_off
    max_val = np.max(I)
    return max_val, x_max, y_max, shift_x, shift_y


# def clean_beam(B, search_box=20):
#     """Clean beam.

#     Function extract the clean beam image from the dirty beam image.
#     A box of size search_box around the peak of the dirty beam image is
#     extracted and padded to the original size of the dirty beam image.

#     Parameters
#     ----------
#     B : np.ndarray
#         The beam image.
#     search_box : int
#         The search box size in pixels.

#     Returns
#     -------
#     B_clean : np.ndarray
#         The cleaned beam image.
#     """
#     max_val, x_max, y_max, shift_x, shift_y = find_peak(B)
#     B_clean = B[y_max-search_box:y_max+search_box,x_max-search_box:x_max+search_box]
#     B_clean = np.pad(B_clean,((y_max-search_box,B.shape[0]-y_max-search_box),(x_max-search_box,B.shape[1]-x_max-search_box)))
#     return B_clean


def clean_hogbom(
    I_obs, B, gamma=0.2, max_iter=100, threshold=None, clean_beam_size_px=2
):
    """Clean Hogbom.

    Function to perform the Hogbom's clean algorithm on a dirty image.

    Parameters
    ----------
    I_obs : np.ndarray
        The dirty image.
    B : np.ndarray
        The beam image (fft shifted).
    gamma : float
        The clean gain.
    max_iter : int
        The maximum number of iterations.
    threshold : float
        The threshold to stop the cleaning process.

    Returns
    -------
    I_clean : np.ndarray
        The cleaned image.
    """
    I_res = I_obs.copy()
    I_clean = np.zeros_like(I_obs)
    sky_model = np.zeros_like(I_obs)
    B_norm = B / np.max(B)
    # B_clean = clean_beam(B_norm, search_box=B_norm.shape[0]//8)
    B_clean = gauss_source(
        B.shape[1], B.shape[0], np.array([0, 0]), fwhm_pix=clean_beam_size_px
    )

    for i in range(max_iter):
        # Get peak coordinates and flux value
        max_val, x_max, y_max, shift_x, shift_y = find_peak(I_res)
        if threshold is not None and max_val < threshold:
            print("Reached threshold at iteration {}".format(i))
            break
        # Subtract the peak from the dirty image
        I_res -= gamma * max_val * shift_beam(B_norm, shift_x, shift_y)
        sky_model[y_max, x_max] += gamma * max_val
        I_clean += gamma * max_val * shift_beam(B_clean, shift_x, shift_y)

    return I_clean + I_res, sky_model



"""Data uils.

This module contains functions to generate synthetic observations for the Argosim project.

:Authors: Ezequiel Centofanti <ezequiel.centofanti@cea.fr>

"""

import numpy as np


def gauss_source(nx=512, ny=512, mu=np.array([0, 0]), sigma=np.eye(2), fwhm_pix=64):
    """Gauss source.

    Function to generate a 2D Gaussian source.

    Parameters
    ----------
    nx : int
        The output image first dimension.
    ny : int
        The output image second dimension.
    mu : np.ndarray
        The mean of the Gaussian source in the [-1,1]x[-1,1] range.
    sigma : np.ndarray
        The covariance matrix of the Gaussian source.
    fwhm_pix : float
        The FWHM of the Gaussian source in pixels.

    Returns
    -------
    source : np.ndarray
        Image of size (nx,ny) containing the 2D Gaussian source.
    """
    fwhm = 2.355
    x, y = np.meshgrid(
        np.linspace(-fwhm / 2, fwhm / 2, nx), np.linspace(-fwhm / 2, fwhm / 2, ny)
    )

    sigma = sigma / (nx * ny) * fwhm_pix**2 / np.sqrt(np.linalg.det(sigma))

    X_unroll = np.array([x.reshape(-1) - mu[0], y.reshape(-1) - mu[1]])
    sigminv = np.linalg.inv(sigma)
    sigminv.dot(X_unroll).shape
    Q = np.sum(np.multiply(X_unroll, sigminv.dot(X_unroll)), axis=0).reshape(nx, ny)
    return np.exp(-Q / 2)  # /(np.sqrt(2*np.pi*np.abs(np.linalg.det(sigma))))


def sigma2d(min_var=5, cov_lim=0.5):
    """Sigma 2D.

    Function to generate a random 2D covariance matrix.

    Parameters
    ----------
    min_var : float
        The minimum variance of both gaussian components.
    cov_lim : float
        The limit of the covariance between gaussian components.

    Returns
    -------
    sigma : np.ndarray
        The 2D covariance matrix.
    """
    var_1 = np.random.rand() + min_var
    # Limit eccentricity
    var_2 = np.random.rand() + min_var
    # Cov <= sqrt(var1 x var2)
    cov12 = (np.random.rand() * 2 - 1) * np.sqrt(var_1 * var_2) * cov_lim
    return np.array([[var_1, cov12], [cov12, var_2]])


def mu2d():
    """Mu 2D.

    Function to generate a random 2D mean vector in the range [-1,1]x[-1,1].

    Returns
    -------
    mu : np.ndarray
        The 2D mean vector.
    """
    return np.random.rand(2) * 2 - 1


def random_source(shape, pix_size):
    """Random source.

    Function to generate 2D Gaussian source with random mean and covariance.

    Parameters
    ----------
    shape : tuple
        The output image shape.
    pix_size : float
        The size in pixels of the Gaussian source.

    Returns
    -------
    source : np.ndarray
        Image of size (nx,ny) containing the 2D Gaussian source.
    """
    mu = mu2d()
    sigma = sigma2d()
    return gauss_source(shape[0], shape[1], mu, sigma, pix_size)


def n_source_sky(shape_px, pix_size_list, source_intensity_list):
    """N source sky.

    Function to generate a sky image with multiple Gaussian sources at random positions.

    Parameters
    ----------
    shape_px : tuple
        The image size in pixels (Nx, Ny).
    pix_size_list : list
        The size in pixels of the Gaussian sources.
    source_intensity_list : list
        The intensity of each Gaussian source in the final image.
        The sum of all the sources should be equal to 1 to have a normalized image.

    Returns
    -------
    sky : np.ndarray
        Image of size (nx,ny) containing the sky model.
    """
    return sum(
        [
            random_source((shape_px[0], shape_px[1]), pix_size) * intensity
            for pix_size, intensity in zip(pix_size_list, source_intensity_list)
        ]
    )



"""Imaging utils.

This module contains functions to perform radio interferometric imaging. 

:Authors: Ezequiel Centofanti <ezequiel.centofanti@cea.fr>

"""

import numpy as np
import numpy.random as rnd
# from PIL import Image

########################################
#           Grid uv samples            #
########################################

# def get_uv_plane(baseline, uv_dim=128):
#     """Get uv plane.

#     Function to compute the uv sampling mask from the baselines list.
#     Perform a 2D histogram of the baselines list with uv_dim bins.

#     Parameters
#     ----------
#     baseline : np.ndarray
#         The baselines list.
#     uv_dim : int
#         The uv-plane sampling mask size.

#     Returns
#     -------
#     uv_plane : np.ndarray
#         The uv sampling mask of the antenna array. The dimensions are (uv_dim, uv_dim).
#         The value of each pixel is the number of uv samples in that pixel.

#     """
#     # Count number of samples per uv grid
#     x_lim=np.max(np.absolute(baseline))#*1.1
#     y_lim=x_lim
#     uv_plane, _, _ = np.histogram2d(baseline[:,0],baseline[:,1],bins=uv_dim, range=[[-x_lim,x_lim],[-y_lim,y_lim]])
#     return np.fliplr(uv_plane.T)#/np.sum(uv_plane, axis=(0,1))

# def get_uv_mask(uv_plane):
#     """Get uv mask.

#     Function to compute the binary mask from the uv sampling grid.

#     Parameters
#     ----------
#     uv_plane : np.ndarray
#         The uv sampling mask.

#     Returns
#     -------
#     uv_plane_mask : np.ndarray
#         The binary mask of the uv sampling mask.
#         The value of each pixel is 1 if the pixel is sampled, 0 otherwise.
#     """
#     # Get binary mask from the uv sampled grid
#     uv_plane_mask = uv_plane.copy()
#     uv_plane_mask[np.where(uv_plane>0)] = 1
#     return uv_plane_mask

# def get_beam(uv_mask):
#     """Get beam.

#     Function to compute the telescope beam from the uv sampling mask.

#     Parameters
#     ----------
#     uv_mask : np.ndarray
#         The uv sampling mask.

#     Returns
#     -------
#     beam : np.ndarray
#         The beam image of the antenna array. The beam is fftshifted (non centered).
#     """
#     return np.abs(np.fft.ifft2(uv_mask))


# def load_sky_model(path):
#     """Load sky model.

#     Function to load a sky model image.

#     Parameters
#     ----------
#     path : str
#         The path to the sky model image.

#     Returns
#     -------
#     sky : np.ndarray
#         The sky model image.
#     """
#     return np.array(Image.open(path).convert("L"))


def sky2uv(sky):
    """Sky to uv plane.

    Function to compute the Fourier transform of the sky.

    Parameters
    ----------
    sky : np.ndarray
        The sky image.

    Returns
    -------
    sky_uv : np.ndarray
        The Fourier transform of the sky.
    """
    # return np.fft.fft2(sky)
    return np.fft.fftshift(np.fft.fft2(np.fft.ifftshift(sky)))


def grid_uv_samples(
    uv_samples, sky_uv_shape, fov_size, mask_type="binary", weights=None
):
    """Grid uv samples.

    Compute the uv sampling mask from the uv samples.

    Parameters
    ----------
    uv_samples : np.ndarray
        The uv samples coordinates in meters.
    sky_uv_shape : tuple
        The shape of the sky model in pixels.
    fov_size : tuple
        The field of view size in degrees.
    mask_type : str
        The type of mask to use. Choose between 'binary', 'histogram' and 'weighted'.
    weights : np.ndarray
        The weights to use for the mask type 'weighted'.

    Returns
    -------
    uv_mask : np.ndarray
        The uv sampling mask.
    uv_samples_indices : np.ndarray
        The indices of the uv samples in pixel coordinates.
    """
    max_u = (180 / np.pi) * sky_uv_shape[0] / (2 * fov_size[0])
    max_v = (180 / np.pi) * sky_uv_shape[1] / (2 * fov_size[1])
    uv_samples_indices = (
        np.rint(uv_samples[:, :2] / np.array([max_u, max_v]) * np.array(sky_uv_shape))
        + np.array(sky_uv_shape) / 2
    )

    if any(np.array(sky_uv_shape) <= np.max(uv_samples_indices, axis=0)):
        raise ValueError(
            "uv samples are out of the uv-plane range. Required Npix > {}".format(
                np.max(uv_samples_indices, axis=0)
            )
        )

    uv_mask = np.zeros(sky_uv_shape, dtype=complex)

    for index in uv_samples_indices:
        if mask_type == "binary":
            uv_mask[int(index[1]), int(index[0])] = 1 + 0j
        elif mask_type == "histogram":
            uv_mask[int(index[1]), int(index[0])] += 1 + 0j
        elif mask_type == "weighted":
            assert (
                weights is not None
            ), "Weights must be provided for mask type 'weighted'."
            uv_mask[int(index[1]), int(index[0])] += weights[
                int(index[0]), int(index[1])
            ]
        else:
            raise ValueError(
                "Invalid mask type. Choose between 'binary', 'histogram' and 'weighted'."
            )

    return uv_mask, uv_samples_indices


def uv2sky(uv):
    """Uv to sky.

    Function to compute the inverse Fourier transform of the uv plane.

    Parameters
    ----------
    uv : np.ndarray
        The image in the uv/Fourier domain.

    Returns
    -------
    sky : np.ndarray
        The image in the sky domain.
    """
    return np.fft.fftshift(np.fft.ifft2(np.fft.ifftshift(uv))).real


def compute_visibilities_grid(sky_uv, uv_mask):
    """Compute visibilities gridded.

    Function to compute the visibilities from the fourier sky and the uv sampling mask.

    Parameters
    ----------
    sky_uv : np.ndarray
        The sky model in Fourier/uv domain.
    uv_mask : np.ndarray
        The uv sampling mask.

    Returns
    -------
    visibilities : np.ndarray
        Gridded visibilities on the uv-plane.
    """
    return sky_uv * uv_mask + 0 + 0.0j


def add_noise_uv(vis, uv_mask, sigma=0.1):
    """Add noise in uv-plane.

    Function to add white gaussian noise to the visibilities in the uv-plane.

    Parameters
    ----------
    vis : np.ndarray
        The visibilities.
    mask : np.ndarray
        The uv sampling mask.
    sigma : float
        The standard deviation of the noise.

    Returns
    -------
    vis : np.ndarray
        The visibilities with added noise.
    """
    noise_sky = rnd.normal(0, sigma, vis.shape)
    noise_uv = sky2uv(noise_sky)

    return vis + compute_visibilities_grid(noise_uv, uv_mask)


# def compute_visibilities(sky_uv, uv_samples_indices):
#     """Compute visibilities.

#     Function to compute the visibilities from the fourier sky and the uv samples.

#     Parameters
#     ----------
#     sky_uv : np.ndarray
#         The sky model in Fourier/uv domain.
#     uv_samples_indices : np.ndarray
#         The indices of the uv samples in pixel coordinates.

#     Returns
#     -------
#     visibilities : np.ndarray
#         List of visibilities.
#     """
#     return np.array([sky_uv[int(v), int(u)] for u,v in uv_samples_indices])

# def get_obs_uv(sky_uv, mask):
#     """Get obs uv.

#     Function to compute the observed uv-plane from the sky model and the uv sampling mask.

#     Parameters
#     ----------
#     sky_uv : np.ndarray
#         The sky model in Fourier/uv domain.
#     mask : np.ndarray
#         The uv sampling mask.

#     Returns
#     -------
#     obs_uv : np.ndarray
#         The observed uv-plane.
#     """
#     return np.fft.fftshift(sky_uv).copy()*mask

# def get_obs_sky(obs_uv, abs=False):
#     """Get obs sky.

#     Function to compute the observed sky model from the observed uv-plane.

#     Parameters
#     ----------
#     obs_uv : np.ndarray
#         The sampled sky model on the uv-plane.
#     abs : bool
#         If True, return the absolute value of the observed sky model.

#     Returns
#     -------
#     obs_sky : np.ndarray
#         The observed sky model.
#     """
#     return np.abs(np.fft.ifft2(np.fft.ifftshift(obs_uv))) if abs else np.fft.ifft2(np.fft.ifftshift(obs_uv))
